// Copyright 2024 Esri
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.


import SwiftUI
import ArcGIS

@MainActor
private class Model: ObservableObject {

    @Published private(set) var offlineMap: Map!

    @Published private(set) var generateOfflineMapJob: GenerateOfflineMapJob!
    private var offlineMapTask: OfflineMapTask!

    @Published private(set) var isGenerateDisabled = true
    @Published var isShowingAlert = false

    private let temporaryDirectory: URL = {
        // swiftlint:disable:next force_try
        return try! FileManager.default.url(
            for: .itemReplacementDirectory,
            in: .userDomainMask,
            appropriateFor: FileManager.default.temporaryDirectory,
            create: true
        )
    }()

    deinit {
        try? FileManager.default.removeItem(at: temporaryDirectory)
    }

    func initializeOfflineMapTask(onlineMap: Map) async {
        do {
            try await onlineMap.load()
            offlineMapTask = OfflineMapTask(onlineMap: onlineMap)
            isGenerateDisabled = false
        } catch {
            print(error)
        }
    }

    private func makeGenerateOfflineMapParameters(areaOfInterest: Envelope) async -> GenerateOfflineMapParameters? {
        do {
            return try await offlineMapTask.makeDefaultGenerateOfflineMapParameters(areaOfInterest: areaOfInterest)
        } catch {
            print(error)
            return nil
        }
    }

    func generateOfflineMap(extent: Envelope) async {
        isGenerateDisabled = true
        guard let parameters = await makeGenerateOfflineMapParameters(areaOfInterest: extent) else {
            isGenerateDisabled = false
            return
        }

        generateOfflineMapJob = offlineMapTask.makeGenerateOfflineMapJob(
            parameters: parameters,
            downloadDirectory: temporaryDirectory
        )
        generateOfflineMapJob.start()

        defer {
            generateOfflineMapJob = nil
            isGenerateDisabled = offlineMap != nil
        }
        do {

            let output = try await generateOfflineMapJob.output
            offlineMap = output.offlineMap
            let builder = EnvelopeBuilder(envelope: extent)
            builder.expand(by: 0.8)
            offlineMap.initialViewpoint = Viewpoint(boundingGeometry: builder.toGeometry())
            isShowingAlert = true

        } catch {
            print(error)
        }

    }

}

struct ContentView: View {

    @StateObject private var model = Model()

    @State private var isGeneratingOfflineMap = false

    @State private var map = Map(
        item: PortalItem(
            portal: .arcGISOnline(connection: .anonymous),

            id: PortalItem.ID("5a030a31e42841a89914bd7c5ecf4d8f")!

        )
    )

    var body: some View {

        GeometryReader { geometry in

            MapViewReader { mapView in

                MapView(map: model.offlineMap ?? map)

                    .interactionModes(isGeneratingOfflineMap ? [] : [.pan, .zoom])

                    .task {
                        await model.initializeOfflineMapTask(onlineMap: map)
                    }

                    .overlay {
                        Rectangle()
                            .stroke(.red, lineWidth: 2)
                            .padding(EdgeInsets(top: 60, leading: 20, bottom: 100, trailing: 20))
                            .opacity(model.offlineMap == nil ? 1 : 0)
                    }

                    .overlay {
                        if isGeneratingOfflineMap,
                           let progress = model.generateOfflineMapJob?.progress {
                            VStack(spacing: 16) {
                                ProgressView(progress)
                                    .progressViewStyle(.linear)
                                    .frame(maxWidth: 200)
                            }
                            .padding()
                            .background(.regularMaterial)
                            .clipShape(RoundedRectangle(cornerRadius: 15))
                            .shadow(radius: 3)
                        }
                    }

                    .alert("Offline map generated", isPresented: $model.isShowingAlert) {
                        Button("Done") {
                            model.isShowingAlert = false
                        }
                    }

                    .toolbar {
                        ToolbarItem(placement: .bottomBar) {

                            Button("Download Map Area") {
                                isGeneratingOfflineMap = true
                            }

                            .disabled(model.isGenerateDisabled || isGeneratingOfflineMap)

                            .task(id: isGeneratingOfflineMap) {
                                guard isGeneratingOfflineMap else { return }
                                let viewRect = geometry.frame(in: .local).inset(
                                    by: UIEdgeInsets(
                                        top: 60,
                                        left: geometry.safeAreaInsets.leading + 20,
                                        bottom: 100,
                                        right: -geometry.safeAreaInsets.trailing + 20
                                    )
                                )
                                guard let extent = mapView.envelope(fromViewRect: viewRect) else { return }
                                await model.generateOfflineMap(extent: extent)
                                isGeneratingOfflineMap = false
                            }

                        }
                    }

            }

        }

    }

}


